/*
 * Name:    mp_net.c
 *
 * Purpose: Python module interface to generic MX net functions.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2006 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MP_NET_DEBUG	FALSE

#include "Python.h"

#include "mx_util.h"
#include "mx_record.h"
#include "mx_net.h"
#include "mx_handle.h"
#include "mx_array.h"
#include "mp.h"

#include "mx_key.h"

/*-------------------------------------------------------------------------*/

static mx_status_type
MpPrivate_net_find_buffer_for_array( MPPRIVATE_NETWORK_FIELD *nf_private )
{
	static const char fname[] = "MpPrivate_net_find_buffer_for_array()";

	/* If you have a 2-dimensional array, the value of the top level
	 * array pointer is the address of the pointer to the first row.
	 * Thus, you have to dereference the top level pointer once to
	 * get to the buffer pointer.
	 *
	 * Similarly, you have to dereference the top level pointer for
	 * a 3-dimensional array twice to get the buffer pointer, three
	 * times for a 4-dimensional array, four times for a 5-dimensional
	 * array, and so forth.
	 */

	union {
		char  *ptr1;
		char **ptr2;
	} buffer;

	long i;

	if ( nf_private == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The nf_private pointer passed was NULL." );
	}

	buffer.ptr1 = nf_private->value_array;

	for ( i = 1; i < nf_private->num_dimensions; i++ ) {

		if ( buffer.ptr1 == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The buffer pointer at i = %ld is NULL.", i );
		}

		/* The value at the pointer should be another pointer. */

		buffer.ptr1 = *buffer.ptr2;
	}

	nf_private->buffer_ptr = buffer.ptr1;

	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

static mx_status_type
MpPrivate_net_setup_field( long nf_handle,
			MX_NETWORK_FIELD **nf )
{
	static const char fname[] = "MpPrivate_net_setup_field()";

	void *nf_pointer;
	MPPRIVATE_NETWORK_FIELD *nf_private;
	long i, datatype, num_dimensions, num_data_element_sizes_to_allocate;
	long dimension_array[MXP_MAX_DIMENSIONS];
	size_t *sizeof_array;
	int new_handle_needed;
	mx_status_type mx_status;

	if ( nf == (MX_NETWORK_FIELD **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_NETWORK_FIELD pointer passed was NULL." );
	}

	mx_status = mx_get_pointer_from_handle( &nf_pointer,
				MpPrivate_global_network_field_handle_table,
				nf_handle );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	*nf = (MX_NETWORK_FIELD *) nf_pointer;

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: nf_handle = %ld, nfname = '%s'",
		fname, nf_handle, (*nf)->nfname));
#endif

	/* Make sure that we have connected to the network field. */

	mx_status = mx_need_to_get_network_handle( *nf, &new_handle_needed );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( new_handle_needed ) {
		mx_status = mx_network_field_connect( *nf );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	/* We need to know the datatype and dimensions of the field.
	 * This information is stored in the MPPRIVATE_NETWORK_FIELD
	 * structure which is pointed to by (*np)->application_ptr
	 */

	if ( (*nf)->application_ptr != NULL ) {

		/* If this pointer is not NULL, then we assume that it
		 * has been set up correctly by a previous call.  If so,
		 * then we can return now.
		 */

		return MX_SUCCESSFUL_RESULT;
	}

	/* If we get here, we need to create and fill in the
	 * MPPRIVATE_NETWORK_FIELD structure.
	 */

	nf_private = malloc( sizeof(MPPRIVATE_NETWORK_FIELD) );

	if ( nf_private == (MPPRIVATE_NETWORK_FIELD *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate an "
			"MPPRIVATE_NETWORK_FIELD structure." );
	}

	(*nf)->application_ptr = nf_private;

	/* Get the field type parameters. */

	mx_status = mx_get_field_type( (*nf)->server_record, (*nf)->nfname,
					MXP_MAX_DIMENSIONS, &datatype,
					&num_dimensions, dimension_array );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Copy the datatype and dimensions to the private field structure. */

	nf_private->datatype = datatype;
	nf_private->num_dimensions = num_dimensions;

	if ( num_dimensions == 0 ) {
		nf_private->dimension_array = NULL;
	} else {
		nf_private->dimension_array
			= malloc( num_dimensions * sizeof(long) );

		if ( nf_private->dimension_array == NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate a %ld element "
			"dimension array for server '%s', field '%s'.",
				num_dimensions,
				(*nf)->server_record->name,
				(*nf)->nfname );
		}
	}

	if ( num_dimensions == 0 ) {
		num_data_element_sizes_to_allocate = 1;
	} else {
		num_data_element_sizes_to_allocate = num_dimensions;
	}

	nf_private->data_element_size_array
		= malloc( num_data_element_sizes_to_allocate * sizeof(size_t) );

	if ( nf_private->data_element_size_array == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %ld element "
		"data element size array for server '%s', field '%s'.",
			num_data_element_sizes_to_allocate,
			(*nf)->server_record->name,
			(*nf)->nfname );
	}

	mx_status = mx_get_datatype_sizeof_array( datatype, &sizeof_array );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Save a copy of the dimension and sizeof arrays. */

	for ( i = 0; i < num_dimensions; i++ ) {
		nf_private->dimension_array[i]         = dimension_array[i];
		nf_private->data_element_size_array[i] = sizeof_array[i];
	}

	/* For a scalar, we always have 1 value in data_element_size_array. */

	if ( num_dimensions == 0 ) {
		nf_private->data_element_size_array[0] = sizeof_array[0];
	}

	/* Allocate an array for the field data to be stored in. */

	if ( num_dimensions == 0 ) {
		nf_private->value_array = malloc( sizeof_array[0] );
	} else {
		nf_private->value_array = mx_allocate_array( num_dimensions,
						dimension_array, sizeof_array );
	}

	if ( nf_private->value_array == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory attempting to allocate a %ld dimensional "
		"array for server '%s', field '%s'.", num_dimensions,
			(*nf)->server_record->name, (*nf)->nfname );
	}

	/* Compute the total size of the data elements in the array. */

	mx_status = mx_get_array_size( nf_private->num_dimensions,
					nf_private->dimension_array,
					nf_private->data_element_size_array,
					&(nf_private->array_size) );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = MpPrivate_net_find_buffer_for_array( nf_private );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: server '%s', field '%s'",
		fname, (*nf)->server_record->name, (*nf)->nfname));

	MX_DEBUG(-2,("%s: datatype = %ld, num_dimensions = %ld",
		fname, nf_private->datatype, nf_private->num_dimensions));

	for ( i = 0; i < nf_private->num_dimensions; i++ ) {
		MX_DEBUG(-2,("%s: dimension_array[%ld] = %ld",
			fname, i, nf_private->dimension_array[i]));
	}

	MX_DEBUG(-2,("%s: array_size = %lu",
		fname, (unsigned long) nf_private->array_size ));

	MX_DEBUG(-2,("%s: value_array = %p, buffer_ptr = %p",
		fname, nf_private->value_array, nf_private->buffer_ptr ));
#endif

	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpPrivate_net_init_cmd( PyObject *self, PyObject *args )
{
	static const char fname[] = "MpPrivate_net_init_cmd()";

	PyObject *result;
	MX_RECORD *server_record;
	long server_handle, new_nf_handle;
	char *record_field_name;
	MX_NETWORK_FIELD *new_nf;
	int parse_status;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "ls",
					&server_handle, &record_field_name );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( server_handle,
							&server_record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	/* Allocate memory for a new MX_NETWORK_FIELD structure. */

	new_nf = malloc( sizeof(MX_NETWORK_FIELD) );

	if ( new_nf == (MX_NETWORK_FIELD *) NULL ) {
		MpPrivateErr_SetString( PyExc_MemoryError,
    "%s: Unable to allocate memory for an MX_NETWORK_FIELD structure.", fname );

		return NULL;
	}

	/* Initialize the contents of the MX_NETWORK_FIELD structure. */

	mx_status = mx_network_field_init( new_nf,
					server_record, record_field_name );

	if ( mx_status.code != MXE_SUCCESS ) {
		free( new_nf );
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	/* Create a handle for the MX_NETWORK_FIELD pointer. */

	mx_status = mx_create_handle( &new_nf_handle, 
				MpPrivate_global_network_field_handle_table,
				new_nf );

	if ( mx_status.code != MXE_SUCCESS ) {
		free( new_nf );
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	/* Send the network field handle back to Python. */

	result = Py_BuildValue( "l", new_nf_handle );

	return result;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpPrivate_net_del_cmd( PyObject *self, PyObject *args )
{
#if MP_NET_DEBUG
	static const char fname[] = "MpPrivate_net_del_cmd()";
#endif

	MX_NETWORK_FIELD *nf;
	MPPRIVATE_NETWORK_FIELD *nf_private;
	long nf_handle;
	int python_status;
	mx_status_type mx_status;

	python_status = PyArg_ParseTuple( args, "l", &nf_handle );

	if ( python_status == 0 )
		return NULL;

	mx_status = MpPrivate_net_setup_field( nf_handle, &nf );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	nf_private = (MPPRIVATE_NETWORK_FIELD *) nf->application_ptr;

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s invoked for network field '%s'.",
		fname, nf->nfname));
#endif

	if ( nf_private != NULL ) {

		/* Free the array that the field data was stored in. */

		if ( nf_private->value_array != NULL ) {
			if ( nf_private->num_dimensions == 0 ) {
				mx_free( nf_private->value_array );
			} else {
				mx_free_array( nf_private->value_array,
					nf_private->num_dimensions,
					nf_private->dimension_array,
					nf_private->data_element_size_array );
			}
		}

		/* Free the other arrays in nf_private. */

		mx_free( nf_private->dimension_array );

		mx_free( nf_private->data_element_size_array );

		/* Now free nf_private itself. */

		mx_free( nf_private );
	}

	/* Finish by freeing the MX_NETWORK_FIELD structure. */

	mx_free( nf );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s completed.", fname));
#endif

	Py_INCREF(Py_None);

	return Py_None;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpPrivate_net_get_field_parameters_cmd( PyObject *self, PyObject *args )
{
	PyObject *result;
	PyObject *object;
	PyObject *dimension_tuple;
	PyObject *data_element_size_tuple;
	MX_NETWORK_FIELD *nf;
	MPPRIVATE_NETWORK_FIELD *nf_private;
	long i, tuple_length, nf_handle;
	int python_status;
	mx_status_type mx_status;

	/* Get the MX_NETWORK_FIELD pointer from the supplied handle. */

	python_status = PyArg_ParseTuple( args, "l", &nf_handle );

	if ( python_status == 0 )
		return NULL;

	mx_status = MpPrivate_net_setup_field( nf_handle, &nf );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	nf_private = (MPPRIVATE_NETWORK_FIELD *) nf->application_ptr;

	/* Create a tuple to contain the dimension sizes. */

	dimension_tuple = PyTuple_New( nf_private->num_dimensions );

	if ( dimension_tuple == NULL )
		return NULL;

	for ( i = 0; i < nf_private->num_dimensions; i++ ) {
		object = PyInt_FromLong( nf_private->dimension_array[i] );

		if ( object == NULL )
			return NULL;

		python_status = PyTuple_SetItem( dimension_tuple, i, object );

		if ( python_status != 0 )
			return NULL;
	}

	/* Create a tuple to contain the data element sizes.  The first
	 * element of the tuple will always be present, even for scalars,
	 * since we still need to know the size of the scalar.
	 */

	if ( nf_private->num_dimensions == 0 ) {
		tuple_length = 1;
	} else {
		tuple_length = nf_private->num_dimensions;
	}

	data_element_size_tuple = PyTuple_New( tuple_length );

	if ( data_element_size_tuple == NULL )
		return NULL;

	for ( i = 0; i < tuple_length; i++ ) {
		object = PyInt_FromLong(nf_private->data_element_size_array[i]);

		if ( object == NULL )
			return NULL;

		python_status = PyTuple_SetItem( data_element_size_tuple,
							i, object );

		if ( python_status != 0 )
			return NULL;
	}

	/* Send the field parameters back to Python. */

	result = Py_BuildValue( "llOOl", nf_private->datatype,
					nf_private->num_dimensions,
					dimension_tuple,
					data_element_size_tuple,
					nf_private->array_size );

	return result;
}

/*-------------------------------------------------------------------------*/

static PyObject *
MpPrivate_net_get( PyObject *self,
		PyObject *args,
		MX_NETWORK_FIELD **nf,
		MPPRIVATE_NETWORK_FIELD **nf_private )
{
#if 0
	static const char fname[] = "MpPrivate_net_get()";
#endif

	long nf_handle;
	int parse_status;
	mx_status_type mx_status;

	/* Get the MX_NETWORK_FIELD pointer from the supplied handle. */

	parse_status = PyArg_ParseTuple( args, "l", &nf_handle );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_net_setup_field( nf_handle, nf );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	*nf_private = (MPPRIVATE_NETWORK_FIELD *) (*nf)->application_ptr;

	/* At this point, MpPrivate_net_setup_field() has already done
	 * everything that the top level mx_get_array() function would
	 * have done, so instead we call mx_internal_get_array() directly.
	 */

	mx_status = mx_internal_get_array( NULL, NULL, (*nf),
					(*nf_private)->datatype,
					(*nf_private)->num_dimensions,
					(*nf_private)->dimension_array,
					(*nf_private)->value_array );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/* The following routine returns the data to Python as a binary string
 * and is intended for use by NumPy.
 */

MX_EXPORT PyObject *
MpPrivate_net_get_binary_string_cmd( PyObject *self, PyObject *args )
{
	PyObject *result;
	PyObject *string_object;
	char *string_buffer;
	MX_NETWORK_FIELD *nf;
	MPPRIVATE_NETWORK_FIELD *nf_private;
	mx_status_type mx_status;

	/* Suppress GCC 4 uninitialized variable warning. */

	nf_private = NULL;

	/* Get the field value into the array nf_private->value_array. */

	result = MpPrivate_net_get( self, args, &nf, &nf_private );

	if ( result == NULL )
		return NULL;

	/*
	 * FIXME: When mx_allocate_array() has been changed to use a flat
	 *        data buffer, it will no longer be necessary to invoke
	 *        mx_copy_array_to_buffer() here.  Then we will be able
	 *        to just pass the pointer for the MX data buffer to
	 *        the function PyString_FromStringAndSize() directly.
	 *
	 *        This will be more efficient since it will effectively
	 *        eliminate one set of buffer copies.
	 */

	/* Create an uninitialized Python string to contain the value(s)
	 * returned by MX.
	 */

	string_object = PyString_FromStringAndSize( NULL,
						nf_private->array_size );

	if ( string_object == NULL )
		return NULL;

	/* Get a pointer to the new string buffer. */

	string_buffer = PyString_AsString( string_object );

	if ( string_buffer == NULL )
		return NULL;

	/* Now copy the MX array to the Python string buffer.
	 *
	 * See the FIXME above for why this should go away some day.
	 */

	if ( nf_private->num_dimensions == 0 ) {

		/* 0-dimensional fields must be handled specially,
		 * since nf_private->dimension_array will be NULL
		 * and mx_copy_array_to_buffer() treats it as an
		 * error if any of its arguments is a NULL pointer.
		 *
		 * However, it is an easy case to handle.
		 */

		memcpy( string_buffer, nf_private->value_array,
						nf_private->array_size );
	} else {
		mx_status = mx_copy_array_to_buffer(
					nf_private->value_array,
					TRUE,
					nf_private->datatype,
					nf_private->num_dimensions,
					nf_private->dimension_array,
					nf_private->data_element_size_array,
					string_buffer,
					nf_private->array_size,
					NULL, FALSE );

		if ( mx_status.code != MXE_SUCCESS )
			return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	return string_object;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/* The following routine returns the data as a potentially multi-dimensional
 * list and is intended for use by plain Python without NumPy.
 */

MX_EXPORT PyObject *
MpPrivate_net_get_object_cmd( PyObject *self, PyObject *args )
{
#if MP_NET_DEBUG
	static const char fname[] = "MpPrivate_net_get_object_cmd()";
#endif

	PyObject *result, *new_sequence;
	MX_NETWORK_FIELD *nf;
	MPPRIVATE_NETWORK_FIELD *nf_private;
	MX_RECORD_FIELD temp_record_field;
	mx_status_type mx_status;

#if 1
	/* This function is not currently complete. */

	return NULL;
#endif

	/* Suppress GCC 4 uninitialized variable warning. */

	nf_private = NULL;

	/* Get the field value into the array nf_private->value_array. */

	result = MpPrivate_net_get( self, args, &nf, &nf_private );

	if ( result == NULL )
		return NULL;

	/* Initialize the temp_record_field structure that we will use
	 * to traverse the array.
	 */

	mx_status = mx_initialize_temp_record_field( &temp_record_field,
					nf_private->datatype,
					nf_private->num_dimensions,
					nf_private->dimension_array,
					nf_private->data_element_size_array,
					nf_private->value_array );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: mx_construct_temp_record_field() succeeded.", fname));
#endif

	/* Now create the list to send back to Python. */

#if 0
	mx_status = MpPrivate_create_sequence_from_field( &temp_record_field,
							MXP_ST_LIST,
							&new_sequence );
	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
#else
	new_sequence = NULL;
#endif

	return new_sequence;
}

/*-------------------------------------------------------------------------*/

/* The following routine receives data from Python as a binary string
 * and is intended for use by NumPy.
 */

MX_EXPORT PyObject *
MpPrivate_net_put_binary_string_cmd( PyObject *self, PyObject *args )
{
#if MP_NET_DEBUG
	static const char fname[] = "MpPrivate_net_put_binary_string_cmd()";
#endif

	MX_NETWORK_FIELD *nf;
	MPPRIVATE_NETWORK_FIELD *nf_private;
	long nf_handle;
	int parse_status, binary_string_length;
	char *binary_string;
	mx_status_type mx_status;

	binary_string = NULL;	/* Suppress GCC 4 warning. */

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: self = %p, args = %p", fname, self, args));
#endif

	parse_status = PyArg_ParseTuple( args, "ls#",
			&nf_handle, &binary_string, &binary_string_length );

	if ( parse_status == 0 )
		return NULL;

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: nf_handle = %ld, binary_string_length = %d",
			fname, nf_handle, binary_string_length));
#endif

	mx_status = MpPrivate_net_setup_field( nf_handle, &nf );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	nf_private = (MPPRIVATE_NETWORK_FIELD *) nf->application_ptr;

	/* For string types, make sure that Python's buffer will fit into
	 * MX's buffer.  For number types, make sure that Python's buffer
	 * and MX's buffer are the same size.
	 */

	switch( nf_private->datatype ) {
	case MXFT_STRING:
		if ( binary_string_length > nf_private->array_size ) {
			MpPrivateErr_SetString( MpPrivateExc_Would_Exceed_Limit,
				"The supplied Python string of length %ld "
				"is too long to fit into MX's buffer "
				"of length %ld for field '%s'.",
					binary_string_length,
					nf_private->array_size, nf->nfname );
			return NULL;
		}
		break;

	case MXFT_RECORD:
	case MXFT_RECORDTYPE:
	case MXFT_INTERFACE:
		MpPrivateErr_SetString( MpPrivateExc_Illegal_Argument,
			"Cannot write to MX field '%s' since writing to "
			"'record', 'recordtype' and 'interface' fields "
			"is not supported.", nf->nfname );
		return NULL;
		break;

	case MXFT_CHAR:
	case MXFT_UCHAR:
	case MXFT_SHORT:
	case MXFT_USHORT:
	case MXFT_BOOL:
	case MXFT_LONG:
	case MXFT_ULONG:
	case MXFT_FLOAT:
	case MXFT_DOUBLE:
	case MXFT_HEX:
	case MXFT_INT64:
	case MXFT_UINT64:
		if ( binary_string_length != nf_private->array_size ) {
			MpPrivateErr_SetString( MpPrivateExc_Type_Mismatch,
				"Python and MX do not agree on the length of "
				"the data buffer for field '%s'.  Python says "
				"that it should be %ld, while MX says that it "
				"should be %ld.  This should never happen.",
					nf->nfname, binary_string_length,
					nf_private->array_size );

			return NULL;
		}
		break;

	default:
		MpPrivateErr_SetString( MpPrivateExc_Unsupported,
			"Cannot write to MX field '%s' since it has an "
			"unsupported datatype of %ld",
				nf->nfname, nf_private->datatype );
		break;
	}

	/* Copy the data in the Python string into the MX value array.
	 *
	 * FIXME: Is there a way of doing this without copying the array?
	 */

	if ( nf_private->num_dimensions == 0 ) {

		/* 0-dimensional fields must be handled specially,
		 * since nf_private->dimension_array will be NULL
		 * and mx_copy_array_to_buffer() treats it as an
		 * error if any of its arguments is a NULL pointer.
		 *
		 * However, it is an easy case to handle.
		 */

		memcpy( nf_private->value_array, binary_string,
						nf_private->array_size );
	} else {
#if 1
		/* For string values, the string length reported by Python
		 * does not include the space taken by the null terminator.
		 */

		if ( nf_private->datatype == MXFT_STRING ) {
			binary_string_length++;
		}
#endif
		mx_status = mx_copy_buffer_to_array(
					binary_string,
					binary_string_length,
					nf_private->value_array,
					TRUE,
					nf_private->datatype,
					nf_private->num_dimensions,
					nf_private->dimension_array,
					nf_private->data_element_size_array,
					NULL, FALSE );

		if ( mx_status.code != MXE_SUCCESS )
			return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	/* At this point, MpPrivate_net_setup_field() has already done
	 * everything that the top level mx_put_array() function would
	 * have done, so instead we call mx_internal_put_array() directly.
	 */

	mx_status = mx_internal_put_array( NULL, NULL, nf,
					nf_private->datatype,
					nf_private->num_dimensions,
					nf_private->dimension_array,
					nf_private->value_array );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: mx_internal_put_array() succeeded.", fname));
#endif

	Py_INCREF(Py_None);

	return Py_None;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/* The following routine receives a sequence from Python and sends it on
 * to MX.  It is intended for use by plain Python without NumPy.
 */

MX_EXPORT PyObject *
MpPrivate_net_put_object_cmd( PyObject *self, PyObject *args )
{
	return NULL;
}

