/*
 * Name:    mp_user_interrupt.c
 *
 * Purpose: Python module interface to generic MX user interrupt functions.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2002, 2004, 2006 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "Python.h"

#include <math.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_handle.h"
#include "mx_key.h"
#include "mx_scan.h"
#include "mp.h"

/*-------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpPrivate_getch_cmd( PyObject *self, PyObject *args )
{
	PyObject *result;
	int character;

	character = mx_getch();

	result = Py_BuildValue( "i", character );

	return result;
}

MX_EXPORT PyObject *
MpPrivate_kbhit_cmd( PyObject *self, PyObject *args )
{
	PyObject *result;
	int return_value;

	return_value = mx_kbhit();

	result = Py_BuildValue( "i", return_value );

	return result;
}

/*-------------------------------------------------------------------------*/

static PyObject *MpPrivate_user_interrupt_function = NULL;

static int MpPrivate_user_interrupt_handler( void );

MX_EXPORT PyObject *
MpPrivate_user_requested_interrupt_cmd( PyObject *self, PyObject *args )
{
	PyObject *result;
	int user_interrupt;

	user_interrupt = mx_user_requested_interrupt();

	result = Py_BuildValue( "i", user_interrupt );

	return result;
}

MX_EXPORT PyObject *
MpPrivate_set_user_interrupt_function_cmd( PyObject *self, PyObject *args )
{
	static const char cname[] = "set_user_interrupt_function";

	PyObject *function_object;
	int parse_status;

	parse_status = PyArg_ParseTuple( args, "O", &function_object );

	if ( parse_status == 0 )
		return NULL;

	/* Get rid of any old function object. */

	if ( MpPrivate_user_interrupt_function != NULL ) {
		Py_DECREF( MpPrivate_user_interrupt_function );

		MpPrivate_user_interrupt_function = NULL;
	}

	/* If the argument passed is None, then change the user interrupt
	 * handler back to the default handler and exit.
	 */

	if ( function_object == Py_None ) {
		mx_set_user_interrupt_function( NULL );

		Py_INCREF(Py_None);
		return Py_None;
	}

	/* Is the argument passed a callable object? */

	if ( PyCallable_Check( function_object ) == 0 ) {
		MpPrivateErr_SetString( PyExc_RuntimeError,
"The argument passed to '%s' is not a function or other callable object.",
			cname );

		return NULL;
	}

	/* Save the supplied function argument. */

	MpPrivate_user_interrupt_function = function_object;

	Py_INCREF( MpPrivate_user_interrupt_function );

	/* Install an MX user interrupt handler. */

	mx_set_user_interrupt_function( MpPrivate_user_interrupt_handler );

	Py_INCREF(Py_None);

	return Py_None;
}

static int
MpPrivate_user_interrupt_handler( void )
{
	static const char fname[] = "MpPrivate_user_interrupt_handler()";

	PyObject *arglist;
	PyObject *result;
	PyTypeObject *result_type;
	int user_interrupt;

	if ( MpPrivate_user_interrupt_function == NULL ) {
		fprintf( stderr, "%s: Attempted to invoke the user interrupt "
		"handler but the MpPrivate_user_interrupt_function pointer "
		"was NULL.", fname );

		return (-1);
	}

	/* Invoke the Python user interrupt handler. */

	arglist = Py_BuildValue("()");

	if ( arglist == NULL ) {
		fprintf( stderr, "%s: arglist was NULL.\n", fname );

		PyErr_Clear();

		return (-1);
	}

	result = PyEval_CallObject( MpPrivate_user_interrupt_function,
					arglist );

	Py_DECREF( arglist );

	if ( result == NULL ) {

#if ( PY_MAJOR_VERSION >= 2 )
		PyErr_WriteUnraisable( MpPrivate_user_interrupt_function );
#endif
		PyErr_Clear();

		return (-1);
	}

	/* The result from the user interrupt handler is interpreted
	 * based on its object type.
	 */

	result_type = (PyTypeObject *) PyObject_Type( result );

	/* If the result was a list, use the first element of the list
	 * as the result.
	 */

	if ( result_type == &PyList_Type ) {
		PyObject *first_list_element;
		int list_size;

		list_size = PyList_Size( result );

		if ( list_size == 0 ) {
			Py_DECREF( result );
			Py_INCREF( Py_None );

			result = Py_None;
		} else {
			first_list_element = PyList_GetItem( result, 0 );

			Py_DECREF( result );
			Py_INCREF( first_list_element );

			result = first_list_element;
		}
	}

	/* Figure out whether or not the external procedure reported
	 * a user interrupt.  We try to handle all types that might
	 * make sense to be used here.
	 */

	if ( result == Py_None ) {

		user_interrupt = 0;
	} else
	if ( result_type == &PyInt_Type ) {

		user_interrupt = PyInt_AsLong( result );
	} else
	if ( result_type == &PyLong_Type ) {

		user_interrupt = PyLong_AsLong( result );
	} else
	if ( result_type == &PyFloat_Type ) {

		user_interrupt = mx_round( PyFloat_AsDouble( result ) );
	} else
	if ( result_type == &PyComplex_Type ) {
		double real_part, imaginary_part, magnitude;

		real_part = PyComplex_RealAsDouble( result );

		imaginary_part = PyComplex_ImagAsDouble( result );

		magnitude = sqrt( real_part * real_part
					+ imaginary_part * imaginary_part );

		if ( magnitude < 0.5 ) {
			user_interrupt = 0;
		} else {
			user_interrupt = 1;
		}
	} else
	if ( result_type == &PyTuple_Type ) {
		int parse_status;

		parse_status = PyArg_ParseTuple( result, "i", &user_interrupt );

		if ( parse_status == 0 ) {
			user_interrupt = -1;
		}
	} else {
		PyObject *procedure_name_repr;
		PyObject *result_repr;

		procedure_name_repr = 
			PyObject_Repr( MpPrivate_user_interrupt_function );

		if ( procedure_name_repr == NULL ) {
			mx_warning(
		"The repr() for the user interrupt function was None.  "
		"This really should not happen." );
		}

		result_repr = PyObject_Repr( result );

		if ( result_repr == NULL ) {
			mx_warning(
		"The repr() for the user interrupt result was None.  "
		"This really should not happen." );
		}

		MpPrivateErr_SetString( PyExc_TypeError,
		"Unsupported result type used for "
		"MP user interrupt handler '%s'.  Result = '%s'",
			PyString_AsString( procedure_name_repr ),
			PyString_AsString( result_repr ) );

		Py_DECREF( procedure_name_repr );
		Py_DECREF( result_repr );

		user_interrupt = (-1);
	}

	Py_DECREF( result );
	Py_DECREF( result_type );

	if ( user_interrupt < 0 ) {

#if ( PY_MAJOR_VERSION >= 2 )
		PyErr_WriteUnraisable( MpPrivate_user_interrupt_function );
#endif
		PyErr_Clear();

		return (-1);
	}

	if ( user_interrupt != 0 ) {
		user_interrupt = 1;
	}

	return user_interrupt;
}

/*-------------------------------------------------------------------------*/

static PyObject *MpPrivate_scan_pause_request_function = NULL;

static mx_status_type MpPrivate_scan_pause_request_handler( MX_SCAN * );

MX_EXPORT PyObject *
MpPrivate_set_scan_pause_request_handler_cmd( PyObject *self, PyObject *args )
{
	static const char cname[] = "set_scan_pause_request_handler";

	PyObject *function_object;
	int parse_status;

	parse_status = PyArg_ParseTuple( args, "O", &function_object );

	if ( parse_status == 0 )
		return NULL;

	/* Get rid of any old function object. */

	if ( MpPrivate_scan_pause_request_function != NULL ) {
		Py_DECREF( MpPrivate_scan_pause_request_function );

		MpPrivate_scan_pause_request_function = NULL;
	}

	/* If the argument passed is None, then change the scan pause
	 * request handler back to the default handler and exit.
	 */

	if ( function_object == Py_None ) {
		mx_set_scan_pause_request_handler( NULL );

		Py_INCREF(Py_None);
		return Py_None;
	}

	/* Is the argument passed a callable object? */

	if ( PyCallable_Check( function_object ) == 0 ) {
		MpPrivateErr_SetString( PyExc_RuntimeError,
"The argument passed to '%s' is not a function or other callable object.",
			cname );

		return NULL;
	}

	/* Save the supplied function argument. */

	MpPrivate_scan_pause_request_function = function_object;

	Py_INCREF( MpPrivate_scan_pause_request_function );

	/* Install an MX scan pause request handler. */

	mx_set_scan_pause_request_handler(MpPrivate_scan_pause_request_handler);

	Py_INCREF(Py_None);

	return Py_None;
}

static mx_status_type
MpPrivate_scan_pause_request_handler( MX_SCAN *scan )
{
	static const char fname[] = "MpPrivate_scan_pause_request_handler()";

	PyObject *arglist;
	PyObject *result;
	PyObject *handler_str_object;
	MX_RECORD *scan_record;
	char *scan_pause_request_handler_name;
	mx_status_type mx_status;

	/* Get the record corresponding to this scan. */

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_SCAN pointer passed was NULL." );
	}

	scan_record = scan->record;

	if ( scan_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	    "The MX_RECORD pointer for the MX_SCAN pointer passed was NULL.");
	}

	if ( MpPrivate_scan_pause_request_function == NULL ) {
		return mx_error( MXE_NOT_AVAILABLE, fname,
		"%s: Attempted to invoke the scan pause request handler but "
		"the MpPrivate_scan_pause_request_function pointer "
		"was NULL.", fname );
	}

	/* Invoke the Python scan pause request handler. */

	arglist = Py_BuildValue("()");

	if ( arglist == NULL ) {
		PyErr_Clear();

		return mx_error( MXE_UNKNOWN_ERROR, fname,
		"The arglist result from Py_BuildValue() was NULL.  "
		"This should not happen." );
	}

	result = PyEval_CallObject( MpPrivate_scan_pause_request_function,
					arglist );

	Py_DECREF( arglist );

	if ( result == NULL ) {

		if ( PyErr_ExceptionMatches( MpPrivateExc_Stop_Requested ) ) {
			PyErr_Clear();

			return mx_error( MXE_STOP_REQUESTED, fname,
		"Scan motors will be stopped at the end of the current step." );
		}

		if ( PyErr_ExceptionMatches( MpPrivateExc_Interrupted ) ) {
			PyErr_Clear();

			return mx_error( MXE_INTERRUPTED, fname,
						"Scan aborted." );
		}

#if ( PY_MAJOR_VERSION >= 2 )
		PyErr_WriteUnraisable( MpPrivate_scan_pause_request_function );
#endif
		PyErr_Clear();

		handler_str_object =
			PyObject_Str( MpPrivate_scan_pause_request_function );

		if ( handler_str_object == NULL ) {
			scan_pause_request_handler_name = NULL;
		} else {
			scan_pause_request_handler_name =
				PyString_AsString( handler_str_object );
		}

		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unexpected exception returned by "
		"scan pause request handler '%s'",
			scan_pause_request_handler_name );

		Py_DECREF( handler_str_object );

		return mx_status;
	}

	return MX_SUCCESSFUL_RESULT;
}

