/*
 * Name:    mp_variable.c
 *
 * Purpose: Python module interface to generic MX variable functions.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 2002-2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "Python.h"

#include "mx_util.h"
#include "mx_record.h"
#include "mx_variable.h"
#include "mx_handle.h"
#include "mp.h"

MX_EXPORT PyObject *
MpPrivate_send_variable_cmd( PyObject *self, PyObject *args )
{
	static const char fname[] = "MpPrivate_send_variable_cmd()";

	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	PyObject *field_value_tuple, *repr_object;
	long handle;
	int parse_status;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "lO",
					&handle, &field_value_tuple );

	if ( parse_status == 0 )
		return NULL;

	/* FIXME: The following should be investigated further.  I had
	 * assumed that everything that was passed to me would be a tuple.
	 * Jim Fait reports that tuples of strings cause Python to crash.
	 */
#if 0
	if ( PyTuple_Check( field_value_tuple ) == 0 ) {
#else
	if ( ( PyTuple_Check( field_value_tuple ) == 0 )
	  && ( PyString_Check( field_value_tuple ) == 0 ) )
	{
#endif
		repr_object = PyObject_Repr( field_value_tuple );

		mx_status = mx_error( MXE_TYPE_MISMATCH, fname,
		"The supplied argument '%s' was not a tuple.",
			PyString_AsString( repr_object ) );

		Py_DECREF( repr_object );

		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	mx_status = mx_find_record_field( record, "value", &record_field );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	mx_status = MpPrivate_create_field_from_tuple( record, record_field,
							field_value_tuple );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	/* Send the variable value to the remote server, if any. */

	mx_status = mx_send_variable( record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

MX_EXPORT PyObject *
MpPrivate_receive_variable_cmd( PyObject *self, PyObject *args )
{
	PyObject *result, *field_value_object;
	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	long handle;
	int parse_status;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "l", &handle );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	mx_status = mx_find_record_field( record, "value", &record_field );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	/* Get the variable value from the remote server, if any. */

	mx_status = mx_receive_variable( record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	/* Construct a tuple to return to Python. */

	mx_status = MpPrivate_create_tuple_from_field( record,
							record_field,
							&field_value_object );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

#if 1
	result = Py_BuildValue( "O", field_value_object );
#else
	/* FIXME - For some reason, the following does not work. */

	result = field_value_object;

	if ( PyTuple_Check( field_value_object ) ) {
		/* Replace a 1-element tuple with the object inside 
		 * the tuple.
		 */

		if ( PyTuple_Size( field_value_object ) == 1 ) {
			result = PyTuple_GetItem( field_value_object, 0 );

			Py_DECREF( field_value_object );
		}
	}
#endif

	return result;
}

